#include "music.h"
#include "mididev.h"

void music::stream::stack::push(unsigned pos, unsigned kertoja)
{
    item i(pos, kertoja);
    pino.push_back(i);
}

unsigned music::stream::stack::pop(unsigned nextpos)
{
    unsigned c = size();
    if(!c)return nextpos;
    unsigned n = pino[--c].kertoja;
    unsigned p = pino[c].alku;
    if(n)
        if(!(pino[c].kertoja = n-1))
        {
            pino.erase(pino.end() - 1);
            return nextpos;
        }
    return p;
}

void music::stream::cachedctrl(int ctrl, int value)
{
    if(owner->streamscache[curchan*128+ctrl] != value)
    {
        owner->streamscache[curchan*128+ctrl] = value;
        if(ctrl == -1) // special
            owner->driver->patch(curchan, instru);
        else
            owner->driver->ctrl(curchan, ctrl, value);
    }
}

void music::stream::updstat()
{
    if(curchan < 0)return;
    
    // use cachedctrl to decrease the amount of traffic.    
    cachedctrl(0, bank);
    cachedctrl(-1, instru); //hack
    cachedctrl(91, reverb*127/99);
    cachedctrl(93, chorus*127/99);
    if(curchan != 9)
        cachedctrl(7, volume*127/99);
    cachedctrl(10, pan*127/99);
}

void music::stream::endnote()
{
    if(noteoffchan >= 0)
    {
        owner->driver->noteoff(noteoffchan, noteoffnot, 0);
        noteoffchan = -1;
        if(curchan >= 0)
        {
            owner->freechan(curchan);
            curchan = -1;
        }
    }
}

void music::stream::startnote(int notenum)
{
    if(instru < 0)
    {
        curchan = 9;
        updstat();
        owner->driver->noteon(noteoffchan=9, noteoffnot=-instru, volume*velocity/99);
        curchan = -1;
    }
    else
    {
        curchan = owner->allochan();
        updstat();
        owner->driver->noteon(noteoffchan=curchan, noteoffnot=notenum, velocity*127/99);
    }
}

unsigned music::stream::readnum(int &dest, unsigned n) const
{
    if(n >= data.size() || data[n] < '0' || data[n] > '9')return 0;
    int value;
    for(value = 0; ; )
    {
        value = value * 10 + data[n] - '0';
        ++n;
        if(n >= data.size() || data[n] < '0' || data[n] > '9')break;
    }
    dest = value;
    return n;
}

bool music::stream::dead()
{
    if(wait || substreams.size())return false;
    return (tilanne == died || posi >= data.size());
}

void music::stream::die(const char *reason)
{
	if(reason)
	{
		fprintf(stderr, "music::stream::die(): \33[31m%s at `", reason);
		data += "'";
		unsigned a, b=data.size();
		for(a=0; a<b; ++a)
		{
			if(a==posi)fprintf(stderr, "\33[1m");
			fprintf(stderr, "%c", data[a]);
			if(a==posi)fprintf(stderr, "\33[21m");
		}
		fprintf(stderr, "\33[m\n");
	}
    stack.clear();
    posi = data.size();
    wait = 0;
    
    if(registered)
    {
	    registered = false;
	    owner->unregisterstream(this);
	}
    substreams.clear();
    endnote();
    tilanne = died;
}

bool music::stream::tick()
{
Redo:
    if(substreams.size())
    {
        unsigned n=0;
        list<stream>::iterator i;
        for(i = substreams.begin(); i != substreams.end(); ++i)
        {
            if(!i->tick())
                ++n;
        }
        if(n) return false;
        
        substreams.clear();
        tilanne = fetch;                        
    }
    if(owner->tickpos % divider)return false;
    if(wait > 0 && --wait)
    {
Return:
        updstat();
        return dead();
    }
    if(tilanne == nuotti)
    {
        endnote();
        tilanne = fetch;
    }
    if(dead())return true;
    if(tilanne == fetch)
        while(!wait && !dead())
        {
        //  fprintf(stderr, "I am \"%s\", executing command: '%c' at posi %u\n",
        //      data.c_str(), data[posi], posi);
        //  fflush(stderr);
            switch(data[posi])
            {
                case '/':
                {
                	unsigned p;
                    if(!(p = readnum(divider, ++posi)))die();
                    posi = p;
                    break;
                }
                case 'p':
                case 'y':
                {
                    bool sammu = data[posi] == 'p';
                    unsigned tmp = readnum(wait, ++posi);
                    if(tmp)
                        posi = tmp;
                    else
                        wait = length;
                    if(!wait)wait = 1;
                    wait *= divider;
                    if(sammu)endnote();
                    break;
                }
                case 'l':
                {
                	unsigned p;
                    if(!(p = readnum(length, ++posi)))die();
                    posi = p;
                    break;
                }
                case 'o':
                {
                	unsigned p;
                    if(!(p = readnum(octave, ++posi)))die();
                    posi = p;
                    break;
                }
                case '<':
                    if(octave > 0)--octave;
                    ++posi;
                    break;
                case '>':
                    if(octave < 7)++octave;
                    ++posi;
                    break;
                case 'a':
                case 'b':
                case 'c':
                case 'd':
                case 'e':
                case 'f':
                case 'g':
                case 'n':
                {
                    int notenum;
                    
                    if(data[posi] == 'n')
                    {
                    	unsigned p;
                        if(!(p = readnum(notenum, ++posi)))
                        {
                        	die();
                        	break;
                        }
                        posi = p;
                        if(data[posi]==':')++posi;
                    }
                    else
                    {
                        static const char n[7] = {9,11,0,2,4,5,7};
                        notenum = n[data[posi++]-'a'] + octave*12;
                        for(; !dead() && (data[posi] == '-' || data[posi]=='+'); ++posi)
                            if(data[posi] == '-') --notenum; else ++notenum;
                    }
                    
                    endnote();
                    startnote(notenum);
                    
                    unsigned tmp = readnum(wait, posi);
                    if(tmp)
                        posi = tmp;
                    else
                        wait = length;
                    if(!wait)wait = 1;
                    wait *= divider;
                    if(data[posi]=='.') ++posi; else tilanne = nuotti;
                    break;
                }
                case 'v':
                {
                    unsigned p;
                    if(!(p = readnum(volume, ++posi)))die();
                    posi = p;
                    break;
                }
                case 'h':
                {
                    unsigned p;
                    if(!(p = readnum(chorus, ++posi)))die();
                    posi = p;
                    break;
                }
                case 'r':
                {
                	unsigned p;
                    if(!(p = readnum(reverb, ++posi)))die();
                    posi = p;
                    break;
                }
                case 'm':
                {
                	unsigned p;
                    if(!(p = readnum(pan, ++posi)))die();
                    posi = p;
                    break;
                }
                case 's':
                {
                	unsigned p;
                    if(!(p = readnum(velocity, ++posi)))die();
                    posi = p;
                    break;
                }
                case 't':
                {
                    unsigned tmp;
                    int a, b;
                    if(!(tmp = readnum(a, posi+1))
                    || data[tmp] != ':'
                    || !(tmp = readnum(b, tmp+1)))
                    	die();
                   	else if(b && (owner->tickpos / divider) % b == (unsigned)a)
	                    posi = tmp;
                    goto Return;
                }
                case 'w':
                {
                    bool accept=false;
                    unsigned tmp=posi;
                    do {
                        int num;
                        if(!(tmp = readnum(num, tmp+1)))
                        {
	                        die();
	                        break;
	                    }
                        if(!accept && owner->isplaying(num))accept = true;
                    } while(data[tmp] == ',');
                    if(accept)posi=tmp;
                    goto Return;
                }
                case '(':
                {
                    int num;
                    unsigned a, b;
                    if(!(a = readnum(num, ++posi)))
                    {
                    	die();
                    	break;
                   	}
                    for(posi=a, b=0; posi < data.size(); ++b, ++posi)
                        if(data[posi] == ')')
                            break;
                    if(posi >= data.size())
                        die();
                    else
                    {
                        ++posi;
                    //  fprintf(stderr, "Defining subproc number %d as '%s'\n", num, data.substr(a,b).c_str());
                    //  fflush(stderr);
                        owner->installstream(num, data.substr(a, b));
                    }
                    break;
                }
                case '{':
                    ++posi;
                    do {
                        unsigned a, b;
                        for(a=posi, b=0; posi < data.size(); ++b, ++posi)
                            if(data[posi] == ',' || data[posi] == '}')
                                break;
                        if(posi >= data.size())
                        {
                            die();
                            break;
                        }
                        stream tmp(data.substr(a, b));
                        tmp.setowner(owner);
                        tmp.restart(*this);
                        substreams.push_back(tmp);
                    } while(data[posi++] == ',');
                    goto Redo;
                case '[':
                {
                    int kertoja = 0;
                    unsigned tmp = readnum(kertoja, ++posi);
                    if(tmp)posi = tmp;
                    stack.push(posi, kertoja);
                    break;
                }
                case ']':
                {
                    posi = stack.pop(posi+1);
                    break;
                }
                case 'x':
                {
                    int num;
                    unsigned p;
                    if(!(p = readnum(num, ++posi)))
                        die();
                    else
                    {
                    	posi = p;
                        stream tmp(owner->program[num]);
                        tmp.setowner(owner);
                        tmp.restart(*this);
                        substreams.push_back(tmp);
                        owner->registerstream(&*substreams.begin(), num);
                    }
                    goto Redo;
                }
                case 'u':
                {
                    int num;
                    unsigned pos2;
                    if(!(pos2 = readnum(num, posi+1)))
                        die();
                    else
                    {
                		data.replace(posi, pos2-posi, owner->program[num]);
                		// Same stream can be registered with many numbers.
                        owner->registerstream(this, num);
                	}
               		goto Redo;
                }
                case 'k':
                {
                    int tmp;
                    unsigned tmp2 = readnum(tmp, ++posi);
                    if(tmp2)
                    {
                        posi = tmp2;
                        owner->killallstreamsnamed(tmp);
                    }
                    else
                        owner->killallstreams();
                    break;
                }
                case 'i':
                {
                    bool miinus = false;
                    if(data[++posi] == '-') { miinus = true; ++posi; }
                    int tmp;
                    unsigned p;
                    if(!(p = readnum(tmp, posi)))
                    {
                    	die();
                    	break;
                    }
                    posi = p;
                    if(miinus)tmp = -tmp;
                    instru = tmp;
                    if(data[posi] != ':')
                        bank = 0;
                    else
                        if(!(posi = readnum(bank, ++posi)))die();
                    updstat();
                    break;
                }
                case ' ':
                case '\t':
                case '\r':
                case '\n':
                    ++posi;
                    break;
                case '\33':
                    if(data[++posi] != '['
                    || data[++posi] != '2'
                    || data[++posi] != 'c')
                        die();
                    else
                        ++posi;
                    break;
                default:
                    die();
            }
        }
    goto Return;
}

music::stream::stream() : tilanne(died), registered(false), noteoffchan(-1)
{
	fprintf(stderr, "Created stream (auto)\n");
}

music::stream::stream(const stream &s)
{
//	fprintf(stderr, "Created stream (assign)\n");
	*this = s;
}

music::stream::stream(const string &s) : registered(false), noteoffchan(-1)
{
//	fprintf(stderr, "Created stream (param)\n");
    // install also sets tilanne (status).
    install(s);
}

music::stream::~stream()
{
    endnote();
//  fprintf(stderr, "Stream: Dying!\n");
    if(registered)
    {
	    registered = false;
	    owner->unregisterstream(this);
	}
}

#include "compress.hh"

void music::upload(const string &s)
{
    if(s[0] == '?')
    {
        string code, arg;
        unsigned p = 1;
        
        while(p < s.size() && isalpha(s[p]))code += s[p++];
        while(p < s.size() && s[p] == ' ')++p;
        
        arg = s.substr(p);
        
        if(code == "list")
        {
            fprintf(stderr, "\n---- \33[1mDefined programs\33[21m ----\n");
            map<int, string>::iterator i;
            
            for(i = program.begin(); i != program.end(); ++i)
            {
                fprintf(stderr, "%5d: %s", i->first, i->second.c_str());
            
                unsigned c=0;
                for(streamlistclass::iterator j = registeredstreams.begin();
                                              j != registeredstreams.end();
                                              ++j)
                    if(j->num == i->first)
                        c++;
                
                if(c)
                {
                    fprintf(stderr, "\33[32m (Running");
                    if(c > 1)fprintf(stderr, " %u times", c);
                    fprintf(stderr, ")\33[m");
                }
                
                fprintf(stderr, "\n");
            }
            fprintf(stderr, "---- \33[1mRunning commands\33[21m ----\n");
            
            list<stream>::iterator j;
            for(j = streams.begin(); j != streams.end(); ++j)
            {
            	fprintf(stderr, "    %s\n", j->data.c_str());
            }
        }
        else if(code == "wipe" || code == "silence")
        {
        	unsigned a;
            list<stream>::iterator i;
            for(a=0, i=streams.begin(); i!=streams.end(); ++i, ++a)i->die("Killed");
            fprintf(stderr, "\n--- \33[1mAll (%u) running streams were killed.\33[21m ----\n", a);
        	if(code == "wipe")
        	{
	            a = program.size();
    	        program.clear();
        	    fprintf(stderr, "--- \33[1mAll (%u) programmed timbres were deleted.\33[21m ----\n", a);
        	}
        }
        else if(code == "undef")
        {
        	int num = strtol(arg.c_str(), NULL, 10);
        	
            map<int, string>::iterator i = program.find(num);
            
            if(i == program.end())
            	fprintf(stderr, "\n---- \33[31m%d was not defined.\33[m ----\n", num);
            else
            {
            	program.erase(i);
            	fprintf(stderr,
            		"\n---- \33[1mProgram %d has been undefined.\33[21m ----"
            		"\n---- Use k command to kill any possible running instances. ----\n",
            		num);
            }
        }
        else if(code == "load")
        {
			FILE *fp = fopen(arg.c_str(), "rb");
			if(!fp)
			{
				fprintf(stderr, "\n--- \33[31mCan't open file `%s'.\33[m ----\n", arg.c_str());
			}
			else
			{
    			fseek(fp, 0, SEEK_END);
    			unsigned size = ftell(fp);
    			char *s = new char[size];
    			if(s)
    			{
    				rewind(fp);
    				unsigned ind = 0;
    				for(;;)
    				{
    					int a = fgetc(fp);
    					if(a == ' ' || a == '\t'
    					|| a == '\r' || a == '\n')continue;
    					if(a == EOF)break;
    					s[ind++] = a;
    				}
    				s[ind] = 0;
    				upload(s);
    				delete[] s;
    			}
    			fclose(fp);
			}
        }
        else if(code == "save")
        {
			FILE *fp = fopen(arg.c_str(), "wb");
			if(!fp)
			{
				fprintf(stderr, "\n--- \33[31mCan't write to file `%s'.\33[m ----\n", arg.c_str());
			}
			else
			{
	            map<int, string>::iterator i;
	            for(i = program.begin(); i != program.end(); ++i)
        	        fprintf(fp, "(%d %s)\n", i->first, compress(i->second).c_str());
				
            	fprintf(stderr,
            		"\n---- \33[1mDefined programs were written to `%s'.\33[21m ----\n",
            		arg.c_str());
    			fclose(fp);
			}
        }
        else
            fprintf(stderr, "\n--- \33[1mInvalid command. Available: ?list, ?wipe, ?load, ?undef, ?silence, ?save\33[21m ----\n");
        
        return;
    }
    stream tmp(s);
    tmp.setowner(this);
    streams.push_back(tmp);
}

void music::registerstream(stream *str, int num)
{
//	fprintf(stderr, "Registering number %d from %p\n", num, (void *)str);
	str->registered = true;
	registeredstreams.push_back(streamlistitem(str, num));
}

void music::unregisterstream(streamlistclass::iterator &i)
{
	i->str->registered = false;
//	fprintf(stderr, "Unregistering stream number %d at %p\n", i->num, (void *)i->str);
	streamlistclass::iterator j = i;
	++j;
	registeredstreams.erase(i);
	i = j;
}

void music::killallstreams()
{
	killallstreamsnamed(-1);
}

void music::unregisterstream(stream *str)
{
//	fprintf(stderr, "Unregistering a stream (size = %u)...\n", registeredstreams.size());
	
	for(streamlistclass::iterator i = registeredstreams.begin();
	                              i != registeredstreams.end(); )
	{
		if(i->str == str)
			unregisterstream(i);
	 	else
	 		++i;
	}
//	fprintf(stderr, "Done unregistering, size now %u\n", registeredstreams.size());
}

void music::killallstreamsnamed(int num)
{
//	fprintf(stderr, "registeredstreams.size() = %u\n", registeredstreams.size());
	for(streamlistclass::iterator i = registeredstreams.begin();
	                              i != registeredstreams.end(); )
	{
		if(num < 0 || i->num == num)
		{
		//	fprintf(stderr, "Killing %u at %p\n", i->num, (void *)i->str);
			i->str->registered = false;
			i->str->die(NULL);
			unregisterstream(i);
		}
		else
			++i;
	}
}

bool music::isplaying(int num)
{
	for(streamlistclass::iterator i = registeredstreams.begin();
	                              i != registeredstreams.end();
	                              ++i)
		if(i->num == num)
			return true;
	return false;
}

music::music() : driver(NULL), tickpos(0)
{
    for(unsigned a=0; a<15; chans[a++] = false);
}
music::music(class mididev *drv) : driver(drv), tickpos(0)
{
    for(unsigned a=0; a<15; chans[a++] = false);
}
music::~music()
{
}
void music::tick()
{
    list<stream>::iterator i;
    for(i=streams.begin(); i!=streams.end(); )
    {
        i->tick();
        if(i->dead())
        {
            list<stream>::iterator j = i;
            ++j;
            streams.erase(i);
            i = j;
        }
        else
            ++i;
    }
    ++tickpos;
//  if(!(tickpos % 100))driver->sleep(1);
}
