#ifndef pokemud_music_h
#define pokemud_music_h

#include <string>
#include <vector>
#include <list>
#include <map>

/* noteoff(ch, note, pressure)
   noteon(ch, note, pressure)
   ctrl(ch, ctrl, value)
   patch(ch, ins)
*/

class music
{
public:
    class mididev *driver;
    unsigned tickpos;
    
    map<unsigned, int> streamscache;
    
private:
    class stream
    {
    private:
        class stack
        {
        public:
            class item
            {
            public:
                unsigned alku, kertoja;
                item(unsigned posi, unsigned n) : alku(posi), kertoja(n) { }
            };
            stack() { }
            ~stack() { }
            void clear() { pino.clear(); }
            unsigned size() const { return pino.size(); }
            void push(unsigned pos, unsigned kertoja);
            unsigned pop(unsigned nextpos);
        private:
            vector<item> pino;
        } stack;
    private:
    	friend music;
        music *owner;
        string data;
        unsigned posi;
        enum { fetch, nuotti, died } tilanne;
        bool registered;
        int wait;
        int divider, length, octave, volume;
        int instru, bank, chorus, reverb, pan;
        int velocity;
        list<stream> substreams;
        int curchan;
        int noteoffchan, noteoffnot;
        
        // returns modified posi (pointing after number) (0 if there was no number);
        unsigned readnum(int &dest, unsigned n) const;
        
        void cachedctrl(int ctrl, int value);        
        void updstat();
        void endnote();
        void startnote(int notenum);
        
    public:
    	stream();
    	stream(const stream &s);
    	stream(const string &s);
        void install(const string &s)
        {
            data = s;
            // sets tilanne
            restart();
        }
        ~stream();
        void setowner(class music *o)
        {
            owner = o;
        }
        void die(const char *reason = "Parse error");
        bool dead();
        
        // gives default attributes
        void restart()
        {
            posi = 0;
            wait = 0;
            tilanne = fetch;
            divider = 4;
            length  = 4;
            octave  = 5;
            volume  = 80;
            instru  = 1;
            bank    = 0;
            chorus  = 10;
            reverb  = 10;
            pan     = 50;
            velocity= 80;
            curchan = -1;
            noteoffchan = -1;
        }
        // gives attributes from parent.
        void restart(const stream &parent)
        {
            restart();
            divider = parent.divider;
            length  = parent.length;
            octave  = parent.octave;
            volume  = parent.volume;
            instru  = parent.instru;        
            bank    = parent.bank;
            chorus  = parent.chorus;
            reverb  = parent.reverb;
            pan     = parent.pan;
            velocity= parent.velocity;
        }
        // returns true if finished, false if still something to do.
        bool tick();
    };
private:
    bool chans[15];
    list<stream> streams;
    
    // To speed up compilation time, we do
    // this and don't include <multimap> here.

    class streamlistitem
    {
    public:
    	stream *str;
    	int num;
    	streamlistitem() : str(NULL), num(0) { }
    	streamlistitem(stream *s, int n) : str(s), num(n) { }
    	~streamlistitem()
    	{
    		// if(str)delete str;
    		/* NO!
    		   This is supposed to be a data
    		   record, not an object collection.
    		*/
    	}
    };
    typedef list<streamlistitem> streamlistclass;
    streamlistclass registeredstreams;
public:
    map<int, string> program;
public:
    int allochan()
    {
        unsigned n;
        for(n=0; n<15; ++n)
            if(!chans[n])
            {
                chans[n] = true;
                return n;
            }
        return 0;
    }
    void freechan(int n)
    {
        if(n > 9)--n;
        if(n > 0 && n < 15)chans[n] = false;
    }
    
    music();
    music(class mididev *drv);
    ~music();
    
    void registerstream(stream *str, int num);
    void unregisterstream(stream *str);
    void unregisterstream(streamlistclass::iterator &i);
    void killallstreams();
    void killallstreamsnamed(int num);
    bool isplaying(int num);
    void installstream(int num, const string &s)
    {
        program[num] = s;
    }
    void setdriver(class mididev *drv)
    {
        driver = drv;
    }
    void upload(const string &s);
    void tick();
};

/* /WWW/midistream-kuvaus-scratch.txt */

#endif
