#include <termio.h>
#include <unistd.h>
#include "music.h"

class terminal
{
    int fd;
    struct termios tattr, saved;
    music *musicdev;
    bool dead;
public:
    terminal() : dead(false)
    {
        tcgetattr(0, &saved);
    //  fcntl(0, F_SETFL, fcntl(0, F_GETFL) | O_NONBLOCK);
        tcgetattr(0, &tattr);
        tattr.c_lflag &= ~(ICANON|ECHO);
        tattr.c_lflag &= ~(ISIG);  // enable ^C
        tattr.c_iflag &= ~(IXON);  // enable ^Q
        tcsetattr(0, TCSAFLUSH, &tattr);
    }
    ~terminal()
    {
        tcsetattr(0, TCSAFLUSH, &saved);
    }
private:
    unsigned char textattr, currentattr;
    unsigned forgive;
    bool justgavedata;
    
    void writefd(const string &buf, int fdee=-1)
    {
        unsigned len = buf.size();
        if(fdee < 0)fdee = fd;
        if(!len)return;
        if(fdee >= 0)
            std::write(fdee, buf.c_str(), buf.size());
    }
    
    void writefd(const char c, int fdee=-1)
    {
        std::write(fdee<0?fd:fdee, &c, 1);
    }
    
    int iansiparm[512], iansipos, iansihit;
    string inputline;
    unsigned inputpos;
    enum ansistate { normal, esc, data, midi };
    ansistate inputstate;
    string midistr;
    
    void textlineout(const string &s)
    {
        unsigned a, b=s.size();
        for(a=0; a<b; ++a)
        {
            char c = s[a];
            if(c == '\\')
            {
                c = s[++a];
                if(c == '\\')
                    writefd(c, 1);
                else
                    setattr(c);
            }
            else
                writefd(c, 1);
        }
    }
    
    void lineout()
    {
        hideoutput();
        textlineout(inputline);
        writefd("\r\n",    1);
        inputline = "";
        inputpos = 0;
        setattr(textattr, true);
    }
    
    // To input undefined ansicode.
    void ansiout(char c)
    {
        anyout("\33[");
        for(int a=0; a<iansipos; ++a)
        {
            if(a)anyout(';');
            int i=iansiparm[a];
            if(!i)
                anyout('0');
            else
            {
                string tmp;
                while(i) { tmp = char('0'+i%10) + tmp; i/=10; }
                anyout(tmp);
            }
        }
        anyout(c);
    }
    void anyout(const string &s)
    {
        unsigned a, b=s.size();
        for(a=0; a<b; anyout(s[a++]));
    }
    void anyout(char c)
    {
        if(c == 8)
        {
            if(inputpos>0)--inputpos;
        }
        else if(c == 13)
            inputpos = 0;
        else if(c == 10)
            lineout();
        else
        {
            bool once = false;
            for(;;)
            {
                if(inputpos < inputline.size())
                    inputline[inputpos++] = c;
                else
                {
                    inputline += c;
                    ++inputpos;
                }
                if(c != '\\' || once)break;
                once = true;
            }
        }
    }
    void setattr(unsigned char c, bool append=false)
    {
        if(append)
        {
            char diu[3] = {'\\', c=='\\'?c^8:c, 0};
            inputline.insert(inputpos, diu);
            inputpos += 2;
            return;
        }
        
        if(currentattr == c)return;
        
        static const char swap[] = "04261537";
        
        string tmp = "\33[";
        bool pp=false, r=false;
        
        if(currentattr == (c|0x08))
            tmp += "21";
        else if(currentattr == (c|0x80))
            tmp += "25";
        else
        {
            if((currentattr&0x80) > (c&0x80)
            || (currentattr&0x08) > (c&0x08))
            {
                tmp += '0'; pp = r = true;
                currentattr = 7;
            }
            if((c & 0x08) && !(currentattr&0x08)) { if(pp)tmp += ';';else pp = true; tmp += '1'; }
            if((c & 0x80) && !(currentattr&0x80)) { if(pp)tmp += ';';else pp = true; tmp += '5'; }
            
            if((c & 0x07) != (currentattr & 0x07) || r) { if(pp)tmp += ';';else pp = true; tmp += '3'; tmp += swap[c&7]; }
            if((c & 0x70) != (currentattr & 0x70) || r) { if(pp)tmp += ';';else pp = true; tmp += '4'; tmp += swap[(c>>4)&7]; }
        }
        tmp += 'm';
        
        writefd(tmp, 1);
        
        currentattr = c;
    }
    string prompt, outputline;
    unsigned outputpos;
    ansistate outputstate;
    int oansiparm[5], oansipos, ohaka;
    bool outputvisible;
    void hideoutput()
    {
        if(!outputvisible)return;
        outputvisible = false;
        
        unsigned p = outputpos;
        unsigned c = outputline.size();
        
        setattr(currentattr & 0x0F);
        
        while(p < c)
        {
            writefd(" ", 1);
            ++p;
        }
        while(p > 0)
        {
            writefd("\b \b", 1);
            --p;
        }
        p = prompt.size();
        while(p > 0)
        {
            writefd("\b \b", 1);
            --p;
        }
    }
    void showoutput()
    {
        if(outputvisible)return;
        outputvisible = true;

        textlineout(prompt);
        
        setattr(7);
        
        unsigned c = outputline.size();
        writefd(outputline, 1);        
        while(c > outputpos)
        {
            writefd("\b", 1);
            --c;
        }
    }
    void fixoutput(unsigned pos, const string &s)
    {
        showoutput();
        if(s == outputline)
        {
            for(; outputpos > pos; --outputpos)writefd("\b", 1);
            writefd(s.substr(outputpos, pos-outputpos), 1);
            outputpos = pos;
            return;
        }
        unsigned p = outputpos;
        unsigned c = outputline.size();
        while(p < c)
        {
            writefd(" ", 1);
            ++p;
        }
        while(p > 0)
        {
            writefd("\b \b", 1);
            --p;
        }
        
        outputline = s;
        outputpos  = pos;
        
        c = outputline.size();
        writefd(outputline, 1);
        while(c > outputpos)
        {
            writefd("\b", 1);
            --c;
        }
    }
    void output()
    {
        char Buf[512], *s;
        int n = std::read(0, Buf, 512);
        if(n <= 0)
        {
            writefd("\n\nEOF from client.\n", 1);
            dead=true;
            return;
        }
        for(s=Buf; n; --n)
        {
            char c = *s++;
            switch(outputstate)
            {
                case midi: // impossible
                case normal:
                    if(c == 27) // esc
                        outputstate = esc;
                    else
                        anykey(c);
                    break;
                case esc:
                    if(c == '[')
                    {
                        outputstate = data;
                        oansiparm[oansipos=0] = 0;
                        ohaka = 0;
                    }
                    else
                    {
                        anykey('\33');
                        anykey(c);
                    }
                    break;
                case data:
                    if(c == ';')
                    {
                        if((unsigned)oansipos < (sizeof oansiparm) / (sizeof(oansiparm[0]))-1)
                            oansiparm[++oansipos] = 0;
                    }
                    else if(c == '[')++ohaka;
                    else if(c >= '0' && c <= '9')
                        oansiparm[oansipos] = oansiparm[oansipos]*10 + c - '0';
                    else
                    {
                        outputstate = normal;
                        switch(c)
                        {
                            case 'A': anykey(''); break;
                            case 'B': anykey(''); break;
                            case 'D': anykey(''); break;
                            case 'C': anykey(''); break;
                            case '~':
                            {
                                switch(oansiparm[0])
                                {
                                    case 1: anykey(''); break;
                                    case 3: anykey(''); break;
                                    case 4: anykey(''); break;
                                }
                                break;
                            }
                        }
                    }
            }
        }
    }
    
    // From keyboard:
    void anykey(unsigned char c);
    
    #if !SIMPLE
    // From socket:
    void input();
    #endif
    
public:
    void run(int fdee, music &midi);
};
